/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
#include "para.h"
#include "close_on_fork.h"
#include "list.h"

#define MAX_STAT_CLIENTS 50

struct stat_client {
	int fd;
	struct list_head node;
};

static struct list_head client_list;
static int initialized;
static int num_clients;



/* This must exactly match para.h */
const char *status_item_list[NUM_STAT_ITEMS] = {
	"status_bar",		"status",		"num_played",
	"mtime",		"length_min",		"length_sec",
	"file_size",		"status_flags",		"format",
	"score",		"audio_file_info1",	"audio_file_info2",
	"audio_file_info3",	"dbinfo1",		"dbinfo2",
	"dbinfo3",		"decoder_flags",	"audiod_status",
	"play_time",		"uptime",		"offset",
	"length",		"stream_start",		"current_time",
	"audiod_uptime",	"dbtool"
};
#define FOR_EACH_STAT_ITEM(i) for (i = 0; i < NUM_STAT_ITEMS; i++)

static void dump_stat_client_list(void)
{
	struct stat_client *sc;

	if (!initialized)
		return;
	list_for_each_entry(sc, &client_list, node)
		PARA_INFO_LOG("stat client on fd %d\n", sc->fd);
}

int stat_client_add(int fd)
{
	struct stat_client *new_client;

	if (num_clients >= MAX_STAT_CLIENTS) {
		PARA_ERROR_LOG("maximal number of stat clients (%d) exceeded\n",
			MAX_STAT_CLIENTS);
		return -1;
	}
	if (!initialized) {
		INIT_LIST_HEAD(&client_list);
		initialized = 1;
	}
	PARA_INFO_LOG("adding client on fd %d\n", fd);
	new_client = para_malloc(sizeof(struct stat_client));
	new_client->fd = fd;
	add_close_on_fork_list(fd);
	list_add(&new_client->node, &client_list);
	dump_stat_client_list();
	num_clients++;
	return 1;
}

void dump_stat_line(char *line)
{
	struct stat_client *sc, *tmp;
	char *buf;
	ssize_t len;

	if (!initialized)
		return;
	buf = make_message("%s\n", line);
	len = strlen(buf);
	list_for_each_entry_safe(sc, tmp, &client_list, node) {
		int fd, ret;
//		PARA_DEBUG_LOG("%s: p=%lx\n", __func__, (long)p);
		fd = sc->fd;
		ret = write(fd, buf, len);
		PARA_DEBUG_LOG("dumped %s to fd %d, ret = %d\n", line, fd, ret);
		if (ret == len )
			continue;
		/* write error */
		close(fd);
		del_close_on_fork_list(fd);
		num_clients--;
		PARA_INFO_LOG("deleting client on fd %d\n", fd);
		list_del(&sc->node);
		free(sc);
		dump_stat_client_list();
	}
	free(buf);
	PARA_DEBUG_LOG("%d client(s)\n", num_clients);
}

void dump_empty_status(void)
{
	int i;
	char *empty_items = NULL;

	if (!initialized)
		return;
	FOR_EACH_STAT_ITEM(i) {
		char *tmp = make_message("%s%s:\n", empty_items? empty_items : "",
			status_item_list[i]);
		free(empty_items);
		empty_items = tmp;
	}
//	PARA_DEBUG_LOG("%s: empty items: %s\n", __func__, empty_items);
	dump_stat_line(empty_items);
	free(empty_items);
}

/*
 * check if line starts with known command.
 */
int stat_line_valid(const char *line)
{
	int i;
	size_t line_len;

	if (!line || !*line)
		return -1;
	line_len = strlen(line);
	for (i = 0; i < NUM_STAT_ITEMS; i++) {
		const char *s = status_item_list[i];
		size_t item_len = strlen(s);

		if (line_len > item_len && line[item_len] == ':' &&
				!strncmp(line, s, item_len))
			return i;
	}
	return -1;
}

/* if line handler is NULL, return number of complete lines in buf.
 * Otherwise, call line_handler for each complete line, but no more than
 * num times. Copy rest of buffer (last chunk containing incomplete line)
 * to start of buffer and return its size.
 */
int check_buf_for_items(char *buf, int n, void (*line_handler)(char *), int num)
{
	char *start = buf, *end;
	int i, num_lines = 0;

	while (start < buf + n) {
		char *next_null;
		char *next_cr;

		next_cr = memchr(start, '\n', buf + n - start);
		next_null = memchr(start, '\0', buf + n - start);
		if (!next_cr && !next_null)
			break;
		if (next_cr && next_null) {
			end = next_cr < next_null? next_cr : next_null;
		} else if (next_null) {
			end = next_null;
		} else
			end = next_cr;
		num_lines++;
		if (line_handler) {
			*end = '\0';
			line_handler(start);
			start = ++end;
			if (num && num_lines >= num)
				break;
		} else
			start = ++end;
	}
	if (!line_handler)
		return num_lines;
	i = buf + n - start;
	if (i && i != n)
		memmove(buf, start, i);
	return i;
}

